/* $Workfile: T2pE1ABCDBit.c $       $Revision: 15 $      $Modtime: 11/06/00 5:21p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2pE1ABCDBit.c

    A simple example console application demonstrating the use of OTX Driver
    to access T1 Rob-Bit signalling bits with the Thor-2-PCI board.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 T1 Lines and expects them to be looped with an external
    loop cable. The T1 lines are used for accessing Bit Robbed Channel
    Associated Signaling information.

    2 Handsets are connected to the codecs and a through connection is
    demonstrated through the Loop Cable.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_DEFAULT_LOG_FILE "T2pE1ABCDBit.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char           *logFileName;
  OTX_INT32       nBoardType;
  int             coLi;
  int             boardNo;
  OTX_BOOL        bChannelizedAbcdBitAccess; // If OTX_TRUE then report ABCD data per channel in the OTX_T1E1_EC_BITROB_SIGNALING event
} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OtxE1ABCDBitDataS  g_sqSendABCDBitData[8];

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Used for Thor-8-PCI (not for Thor-2-PCI)
  OTX_HANDLE hLi[8];
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;


struct {
  OTX_HANDLE hT1E1Cas[8];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI

  // Connect to the Physical Thor-2 Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI and Thor-2-PCMCIA boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
    // On the Thor-2-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.bChannelizedAbcdBitAccess) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_CHANNELIZED_ABCD_BIT_ACCESS, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }

}



/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create CAS Sender and Receiver logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_E1_ABCD_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1Cas[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(g_hPipes.hLiOut[i], result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    ts = 24;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);

  }
}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  return;
}


/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  // Enable Logical T1/E1 Cas Devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hT1E1Cas[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hT1E1Cas[i], result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0xC00000;

  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OtxDrvReleaseLib();
			OsaConsoleExit();
      exit(1);
    }

  }
}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("E1/T1[%d] No string for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (pEventData->m_nParam) {
      DemoLogf("E1/T1[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
    }

    break;

  case OTX_LDEVICE_E1_ABCD_BIT_ACCESS:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("ABCD Device[%d] No String for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }

    if (pEventData->m_nCode == OTX_T1E1_EC_BITROB_SIGNALING) {
      if (g_demoEnv.bChannelizedAbcdBitAccess) {
        DemoLogf("ABCD Device[%d] Event: %s (Channel:%.2d ABCD:0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam >> 8, pEventData->m_nParam & 0x0F);
      } else {
        DemoLogf("ABCD Device[%d] %c Bits changed (0x%0.8x)\n", pEventData->m_nSourceId, 'A' + (pEventData->m_nParam >> 30), pEventData->m_nParam & 0x3fffffff);
      }
    } else {

      if (pEventData->m_nParam) {
        DemoLogf("ABCD Device[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("ABCD Device[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    }
    break;


  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************
  @func
    Extract a specific channel ABCD bit value from the OtxE1ABCDBitDataS struct.
*******************************************************************************/
char *DemoGetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxE1ABCDBitDataS  *pABCDBitData,
  OUT char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  pszValue[0] = '\0';
  
  if (nCasCh < 1 || nCasCh > 30) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chA >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chB >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chC >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chD >> (nCasCh - 1)) & 0x01 ? 1 : 0);

  return pszValue;
}


/*******************************************************************************
  @func
    Sets a specific CAS-channel ABCD bit value in the OtxE1ABCDBitDataS struct.
*******************************************************************************/
char *DemoSetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxE1ABCDBitDataS *pABCDBitData,
  IN char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  OTX_INT32 i;
  OTX_UINT32 *pChX = &(pABCDBitData->chA);
  
  if (nCasCh < 1 || nCasCh > 30) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  for (i = 0; i < 4; i++) {
    switch (i) {
    case 0:
      pChX = &(pABCDBitData->chA);
      break;
    case 1:
      pChX = &(pABCDBitData->chB);
      break;
    case 2:
      pChX = &(pABCDBitData->chC);
      break;
    case 3:
      pChX = &(pABCDBitData->chD);
      break;
    }

    switch (pszValue[i]) {
    case '1':
      *pChX |=   0x01 << (nCasCh - 1);
      break;
    case '0':
      *pChX &= ~(0x01 << (nCasCh - 1));
      break;
    default:
      DemoLogf("Error: Invalid bit value (%c). Expected '0' or '1'\n", pszValue[i]);
      return pszValue;
    }
  }

  return pszValue;
}
  

/*******************************************************************************
  @func
    Prints the ABCD Bit data for all 30 CAS channels
*******************************************************************************/
void DemoPrintAllCasChannelsData(
  OTX_INT32 nCasDev
)
{
  OTX_RESULT result;

  DemoLogf("Reading received bitrob data on LI#%d ... \n", nCasDev);

  if (g_demoEnv.bChannelizedAbcdBitAccess) {
    OTX_INT32 i;
    OTX_UINT8 nABCDBits;

    for (i = 1; i <= 30; i++) {
      result = OtxE1GetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCasDev], i, &nABCDBits); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCasDev], result);
      DemoLogf("   Channel %.2d: 0x%X", i, nABCDBits);
      if (!(i % 2)) {
        DemoLogf("\n");
      }
    }

  } else {
    OTX_INT32 i;
    OtxE1ABCDBitDataS  rxABCDBitData;
    char szABCDbuf[4+1];  // +1 for the '\0'
    
    result = OtxE1GetABCDBit(g_hLogicDevs.hT1E1Cas[nCasDev], &rxABCDBitData); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCasDev], result);
    if (result != OTX_S_OK) {
      return;
    }
    DemoLogf("A = %.6X, B = %.6X, C = %.6X, D = %.6X\n", rxABCDBitData.chA, rxABCDBitData.chB, rxABCDBitData.chC, rxABCDBitData.chD);
    DemoLogf("\n");
    DemoLogf("            ABCD              ABCD\n");
    DemoLogf("            ====              ====\n");
    for (i = 1; i <= 30; i++) {
      DemoLogf("Channel %.2d: ", i);
      DemoLogf("%s", DemoGetCasChannelValue(i, &rxABCDBitData, szABCDbuf));
      DemoLogf("  ");
      if (!(i % 2)) {
        DemoLogf("\n");
      }
    }
    DemoLogf("\n");
  } 
}

/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentCasChannel = 1;
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" n   Increment current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" N   Decrement current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" z   Set transmitted ABCD-bits to ABCD = 0101 on all CAS channels on current Li\n");
    DemoLogf(" s   Query for transmitted ABCD-bits for current CAS channel on current Li\n");
    DemoLogf(" r   Read the value of the received ABCD bit data on the other Li\n");
    DemoLogf(" R   Read the value of the received ABCD bit data on the current Li\n");
    DemoLogf(" l   Query Line Status of current Li\n");
    break;

    case 'z':
      {
        DemoLogf("Setting transmitted ABCD bit data (for all channels) to 0101 (0x55) on LI#%d \n", nCurrentLi);
        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_INT32 i;
          for (i = 1; i <= 30; i++) {
            result = OtxE1SetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], i, 0x05); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCurrentLi], result);
          }
        } else {

          g_sqSendABCDBitData[nCurrentLi].chA = 0x00000000;
          g_sqSendABCDBitData[nCurrentLi].chB = 0xffffffff;
          g_sqSendABCDBitData[nCurrentLi].chC = 0x00000000;
          g_sqSendABCDBitData[nCurrentLi].chD = 0xffffffff;
          result = OtxE1SetABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendABCDBitData[nCurrentLi]); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCurrentLi], result);
        }
        DemoLogf("Done\n");
      }
      break;


    case 's':
      {
        char szABCDbuf[4+1];  // +1 for the '\0'
        char szBuf[12];  // +1 for the '\0'

        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_UINT8 nABCDBits;

          DemoLogf("Enter ABCD-bit data (0x0 - 0xF) for CAS channel %d on Li%d: ", nCurrentCasChannel, nCurrentLi);
          OsaConsoleGetStr(szBuf);
          nABCDBits = (OTX_UINT8)strtoul(szBuf, NULL, 0);
          result = OtxE1SetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], nCurrentCasChannel, nABCDBits); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCurrentLi], result);

        } else {
          DemoGetCasChannelValue(nCurrentCasChannel, &g_sqSendABCDBitData[nCurrentLi], szABCDbuf);
          
          DemoLogf("Enter ABCD-bit data for CAS channel %d on Li%d (ABCD=%s): ", nCurrentCasChannel, nCurrentLi, szABCDbuf);
          OsaConsoleGetStr(szBuf);
          if (*szBuf) {
            if (strlen(szBuf) != 4) {
              DemoLogf("Invalid entry (%s). Please enter ABCD-bit data as a 4 character string.\n", szBuf);
              DemoLogf("E.g. \"0000\", or \"0101\" \n");
              break;
            } else {
              DemoSetCasChannelValue(nCurrentCasChannel, &g_sqSendABCDBitData[nCurrentLi], szBuf);
              result = OtxE1SetABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendABCDBitData[nCurrentLi]); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCurrentLi], result);
            }
          } else {
            DemoLogf("ABCD-bit data remains unchanged for this channel\n");
          }
          DemoLogf("Done.\n");
        }
      }
      break;


    case 'r':
      {
        OTX_INT32 nLi = nCurrentLi + 1;
        
        if ((nCurrentLi % 2) == 0) {
          // Even Li
          nLi = nCurrentLi + 1;
        } else {
          // Odd Li
          nLi = nCurrentLi - 1;
        }
        DemoPrintAllCasChannelsData(nLi);
      }
      break;

    case 'R':
      {
        DemoPrintAllCasChannelsData(nCurrentLi);
      }
      break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      DemoLogf("Line status of Li #%d is ", nCurrentLi);

      result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nCurrentLi], &eLayer1Status); //DemoCheckRc(result);

      DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
    }
    break;

  case 'n':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel++;
      if (nCurrentCasChannel > 30) {
        // Wrap 
        nCurrentCasChannel = 1;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel--;
      if (nCurrentCasChannel < 1) {
        // Wrap
        nCurrentCasChannel = 30;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
    }
    break;
    
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';

      if (nLi > g_demoEnv.coLi) {
        DemoLogf("Error: LI#%d is not supported by this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("LI#%d Selected ....\n", nCurrentLi);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName               = DEMO_DEFAULT_LOG_FILE;
  envP->boardNo                   = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType                = OTX_DEVICE_THOR_PCI;
  envP->coLi                      = 2;
  envP->bChannelizedAbcdBitAccess = OTX_TRUE;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'U') {   // Un-Channelized ABCD bit access
        envP->bChannelizedAbcdBitAccess = OTX_FALSE;
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == '8') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          envP->coLi = 8;
        } else if ( *(argv[i]+2) == 'M') {
          envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          envP->coLi = 2;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
        
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

}


/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[2|8|M]    : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board (default)\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /U           : Use un-channelized ABCD bit access\n");
}


/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 i;

	OsaConsoleInit();

  DemoLogf("T2pE1ABCDBit, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Example program for accessing E1 ABCD-bit data\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("ABCD Bit access is %s\n", g_demoEnv.bChannelizedAbcdBitAccess ? "channelized" : "non-channelized");
  DemoLogf("\nPress 'h' for help\n");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    g_sqSendABCDBitData[i].chA = 0x00000000;
    g_sqSendABCDBitData[i].chB = 0xffffffff;
    g_sqSendABCDBitData[i].chC = 0x00000000;
    g_sqSendABCDBitData[i].chD = 0xffffffff;
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  DemoLogf("Lib connected okay\n");

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

